"""Tests for all pre-built manifests."""
import importlib
import pytest

CONNECTORS = [
    "banking_aa", "gstn", "netsuite", "oracle_fusion", "quickbooks", "sap",
    "stripe", "tally", "zoho_books", "pinelabs_plural", "income_tax_india",
    "darwinbox", "docusign", "epfo", "greenhouse", "keka", "linkedin_talent",
    "okta", "zoom", "salesforce", "hubspot", "mailchimp", "google_ads",
    "meta_ads", "linkedin_ads", "ga4", "mixpanel", "moengage", "ahrefs",
    "bombora", "brandwatch", "buffer", "g2", "trustradius", "wordpress",
    "jira", "confluence", "servicenow", "zendesk", "pagerduty",
    "sanctions_api", "mca_portal", "gmail", "slack", "github",
    "google_calendar", "s3", "sendgrid", "twilio", "twitter", "whatsapp",
    "youtube", "langsmith",
]

VALID_PERMISSIONS = {"read", "write", "delete", "admin"}


@pytest.mark.parametrize("connector", CONNECTORS)
def test_manifest_loads(connector: str) -> None:
    mod = importlib.import_module(f"grantex.manifests.{connector}")
    assert hasattr(mod, "manifest")


@pytest.mark.parametrize("connector", CONNECTORS)
def test_manifest_has_correct_connector_name(connector: str) -> None:
    mod = importlib.import_module(f"grantex.manifests.{connector}")
    # Connector name should match filename (with underscores/hyphens)
    assert mod.manifest.connector == connector.replace("_", "_")


@pytest.mark.parametrize("connector", CONNECTORS)
def test_manifest_has_tools(connector: str) -> None:
    mod = importlib.import_module(f"grantex.manifests.{connector}")
    assert mod.manifest.tool_count >= 1


@pytest.mark.parametrize("connector", CONNECTORS)
def test_manifest_tools_have_valid_permissions(connector: str) -> None:
    mod = importlib.import_module(f"grantex.manifests.{connector}")
    for tool_name, perm in mod.manifest.tools.items():
        assert perm in VALID_PERMISSIONS, f"{connector}.{tool_name} has invalid permission: {perm}"


def test_total_manifest_count() -> None:
    assert len(CONNECTORS) == 53



"""Grantex Python SDK — delegated authorization protocol for AI agents."""

from __future__ import annotations

from ._client import Grantex
from ._errors import (
    GrantexApiError,
    GrantexAuthError,
    GrantexError,
    GrantexNetworkError,
    GrantexTokenError,
)
from ._types import (
    RateLimit,
    Agent,
    RotateKeyResponse,
    SignupParams,
    SignupResponse,
    Anomaly,
    AuditEntry,
    AuthorizationRequest,
    AuthorizeParams,
    ChainIntegrity,
    DetectAnomaliesResponse,
    ListAnomaliesResponse,
    CheckoutResponse,
    ComplianceAuditExport,
    ComplianceExportAuditParams,
    ComplianceExportGrantsParams,
    ComplianceGrantsExport,
    ComplianceSummary,
    EvidencePack,
    EvidencePackMeta,
    EvidencePackParams,
    CreateCheckoutParams,
    CreatePortalParams,
    CreateWebhookParams,
    DelegateParams,
    ExchangeTokenParams,
    ExchangeTokenResponse,
    RefreshTokenParams,
    CreatePrincipalSessionParams,
    PrincipalSessionResponse,
    Grant,
    GrantTokenPayload,
    ListAgentsResponse,
    ListAuditParams,
    ListAuditResponse,
    ListGrantsParams,
    ListGrantsResponse,
    ListWebhooksResponse,
    LogAuditParams,
    CreatePolicyParams,
    ListPoliciesResponse,
    Policy,
    PortalResponse,
    SubscriptionStatus,
    UpdatePolicyParams,
    VerifiedGrant,
    VerifyGrantTokenOptions,
    VerifyTokenResponse,
    WebhookEndpoint,
    WebhookEndpointWithSecret,
    # SCIM
    ScimEmail,
    ScimUserMeta,
    ScimUser,
    ScimListResponse,
    CreateScimUserParams,
    ScimToken,
    ScimTokenWithSecret,
    ListScimTokensResponse,
    # SSO
    SsoConfig,
    CreateSsoConfigParams,
    SsoLoginResponse,
    SsoCallbackResponse,
    # Enterprise SSO
    SsoConnection,
    CreateSsoConnectionParams,
    UpdateSsoConnectionParams,
    ListSsoConnectionsResponse,
    SsoConnectionTestResult,
    SsoEnforcementParams,
    SsoEnforcementResponse,
    SsoSession,
    ListSsoSessionsResponse,
    SsoOidcCallbackParams,
    SsoSamlCallbackParams,
    SsoLdapCallbackParams,
    SsoCallbackResult,
    # Budgets
    AllocateBudgetParams,
    BudgetAllocation,
    DebitBudgetParams,
    DebitBudgetResponse,
    BudgetTransaction,
    BudgetTransactionsResponse,
    # Vault
    StoreCredentialParams,
    StoreCredentialResponse,
    VaultCredential,
    ListVaultCredentialsParams,
    ListVaultCredentialsResponse,
    ExchangeCredentialParams,
    ExchangeCredentialResponse,
    # WebAuthn / FIDO
    WebAuthnRegistrationOptions,
    WebAuthnRegistrationVerifyParams,
    WebAuthnCredential,
    ListWebAuthnCredentialsResponse,
    # Verifiable Credentials
    VerifiableCredentialRecord,
    ListCredentialsParams,
    ListCredentialsResponse,
    VCVerificationResult,
    # SD-JWT
    SDJWTPresentParams,
    SDJWTPresentResult,
    # Developer Settings
    UpdateDeveloperSettingsParams,
    UpdateDeveloperSettingsResponse,
    # Passports (MPP Agent Identity)
    MaxTransactionAmount,
    IssuePassportParams,
    IssuedPassportResponse,
    GetPassportResponse,
    RevokePassportResponse,
    ListPassportsParams,
    ListPassportsResponse,
)
from .resources._passports import PassportsClient
from .resources._events import EventsClient, GrantexEvent as GrantexStreamEvent, StreamOptions, Subscription
from ._pkce import PkceChallenge, generate_pkce
from ._verify import verify_grant_token
from ._webhook import verify_webhook_signature
from .manifest import ToolManifest, Permission, EnforceResult
from ._fastapi import GrantexEnforcer

__version__ = "0.1.6"

__all__ = [
    # Main client
    "Grantex",
    # Signup
    "SignupParams",
    "SignupResponse",
    "RotateKeyResponse",
    # PKCE
    "PkceChallenge",
    "generate_pkce",
    # Standalone verify
    "verify_grant_token",
    # Webhook signature verification
    "verify_webhook_signature",
    # Errors
    "GrantexError",
    "GrantexApiError",
    "GrantexAuthError",
    "GrantexTokenError",
    "GrantexNetworkError",
    # Rate Limits
    "RateLimit",
    # Types
    "Agent",
    "Anomaly",
    "AuditEntry",
    "DetectAnomaliesResponse",
    "ListAnomaliesResponse",
    "AuthorizationRequest",
    "AuthorizeParams",
    "ChainIntegrity",
    "CheckoutResponse",
    "ComplianceAuditExport",
    "EvidencePack",
    "EvidencePackMeta",
    "EvidencePackParams",
    "ComplianceExportAuditParams",
    "ComplianceExportGrantsParams",
    "ComplianceGrantsExport",
    "ComplianceSummary",
    "CreateCheckoutParams",
    "CreatePolicyParams",
    "CreatePortalParams",
    "CreateWebhookParams",
    "DelegateParams",
    "ExchangeTokenParams",
    "ExchangeTokenResponse",
    "RefreshTokenParams",
    "CreatePrincipalSessionParams",
    "PrincipalSessionResponse",
    "Grant",
    "GrantTokenPayload",
    "ListAgentsResponse",
    "ListAuditParams",
    "ListAuditResponse",
    "ListGrantsParams",
    "ListGrantsResponse",
    "ListPoliciesResponse",
    "ListWebhooksResponse",
    "LogAuditParams",
    "Policy",
    "PortalResponse",
    "SubscriptionStatus",
    "VerifiedGrant",
    "VerifyGrantTokenOptions",
    "UpdatePolicyParams",
    "VerifyTokenResponse",
    "WebhookEndpoint",
    "WebhookEndpointWithSecret",
    # SCIM
    "ScimEmail",
    "ScimUserMeta",
    "ScimUser",
    "ScimListResponse",
    "CreateScimUserParams",
    "ScimToken",
    "ScimTokenWithSecret",
    "ListScimTokensResponse",
    # SSO
    "SsoConfig",
    "CreateSsoConfigParams",
    "SsoLoginResponse",
    "SsoCallbackResponse",
    # Enterprise SSO
    "SsoConnection",
    "CreateSsoConnectionParams",
    "UpdateSsoConnectionParams",
    "ListSsoConnectionsResponse",
    "SsoConnectionTestResult",
    "SsoEnforcementParams",
    "SsoEnforcementResponse",
    "SsoSession",
    "ListSsoSessionsResponse",
    "SsoOidcCallbackParams",
    "SsoSamlCallbackParams",
    "SsoLdapCallbackParams",
    "SsoCallbackResult",
    # Budgets
    "AllocateBudgetParams",
    "BudgetAllocation",
    "DebitBudgetParams",
    "DebitBudgetResponse",
    "BudgetTransaction",
    "BudgetTransactionsResponse",
    # Vault
    "StoreCredentialParams",
    "StoreCredentialResponse",
    "VaultCredential",
    "ListVaultCredentialsParams",
    "ListVaultCredentialsResponse",
    "ExchangeCredentialParams",
    "ExchangeCredentialResponse",
    # WebAuthn / FIDO
    "WebAuthnRegistrationOptions",
    "WebAuthnRegistrationVerifyParams",
    "WebAuthnCredential",
    "ListWebAuthnCredentialsResponse",
    # Verifiable Credentials
    "VerifiableCredentialRecord",
    "ListCredentialsParams",
    "ListCredentialsResponse",
    "VCVerificationResult",
    # SD-JWT
    "SDJWTPresentParams",
    "SDJWTPresentResult",
    # Developer Settings
    "UpdateDeveloperSettingsParams",
    "UpdateDeveloperSettingsResponse",
    # Passports (MPP Agent Identity)
    "PassportsClient",
    "MaxTransactionAmount",
    "IssuePassportParams",
    "IssuedPassportResponse",
    "GetPassportResponse",
    "RevokePassportResponse",
    "ListPassportsParams",
    "ListPassportsResponse",
    # Events
    "EventsClient",
    "GrantexStreamEvent",
    "StreamOptions",
    "Subscription",
    # Scope Enforcement / Manifests
    "ToolManifest",
    "Permission",
    "EnforceResult",
    # FastAPI Integration
    "GrantexEnforcer",
    # Version
    "__version__",
]



from __future__ import annotations

from .._http import HttpClient
from .._types import (
    GetPassportResponse,
    IssuedPassportResponse,
    IssuePassportParams,
    ListPassportsParams,
    ListPassportsResponse,
    RevokePassportResponse,
)


class PassportsClient:
    def __init__(self, http: HttpClient) -> None:
        self._http = http

    def issue(self, params: IssuePassportParams) -> IssuedPassportResponse:
        data = self._http.post("/v1/passport/issue", params.to_dict())
        return IssuedPassportResponse.from_dict(data)

    def get(self, passport_id: str) -> GetPassportResponse:
        from urllib.parse import quote

        data = self._http.get(f"/v1/passport/{quote(passport_id, safe='')}")
        return GetPassportResponse.from_dict(data)

    def revoke(self, passport_id: str) -> RevokePassportResponse:
        from urllib.parse import quote

        data = self._http.post(
            f"/v1/passport/{quote(passport_id, safe='')}/revoke"
        )
        return RevokePassportResponse.from_dict(data)

    def list(
        self, params: ListPassportsParams | None = None
    ) -> ListPassportsResponse:
        path = "/v1/passports"
        if params is not None:
            query = params.to_query()
            if query:
                from urllib.parse import urlencode

                path = f"{path}?{urlencode(query)}"
        data = self._http.get(path)
        # Server returns a bare JSON array
        return ListPassportsResponse.from_list(data)



import threading
import time
from unittest.mock import patch

import pytest
from grantex.resources._events import EventsClient, GrantexEvent, StreamOptions, Subscription


class TestGrantexEvent:
    def test_from_dict(self):
        d = {"id": "evt_1", "type": "grant.created", "createdAt": "2026-03-01T00:00:00Z", "data": {"grantId": "grnt_1"}}
        event = GrantexEvent.from_dict(d)
        assert event.id == "evt_1"
        assert event.type == "grant.created"
        assert event.created_at == "2026-03-01T00:00:00Z"
        assert event.data == {"grantId": "grnt_1"}

    def test_from_dict_missing_data(self):
        d = {"id": "evt_1", "type": "grant.created", "createdAt": "2026-03-01T00:00:00Z"}
        event = GrantexEvent.from_dict(d)
        assert event.data == {}

    def test_frozen(self):
        event = GrantexEvent(id="evt_1", type="grant.created", created_at="2026-03-01T00:00:00Z", data={})
        with pytest.raises(AttributeError):
            event.id = "other"


class TestStreamOptions:
    def test_defaults(self):
        opts = StreamOptions()
        assert opts.types is None

    def test_with_types(self):
        opts = StreamOptions(types=["grant.created", "token.issued"])
        assert opts.types == ["grant.created", "token.issued"]


def _make_events(count: int) -> list[GrantexEvent]:
    """Helper to create a list of test events."""
    return [
        GrantexEvent(
            id=f"evt_{i}",
            type="grant.created",
            created_at="2026-03-01T00:00:00Z",
            data={"n": i},
        )
        for i in range(count)
    ]


class TestEventsClient:
    def test_constructor(self):
        client = EventsClient("https://api.grantex.dev", "test-key")
        assert client._base_url == "https://api.grantex.dev"
        assert client._api_key == "test-key"


class TestSubscription:
    def test_active_property(self):
        stop = threading.Event()
        stop.set()  # thread will exit immediately
        t = threading.Thread(target=lambda: None, daemon=True)
        t.start()
        t.join()
        sub = Subscription(thread=t, stop_event=stop)
        assert sub.active is False

    def test_unsubscribe_sets_stop(self):
        stop = threading.Event()
        t = threading.Thread(target=lambda: stop.wait(), daemon=True)
        t.start()
        sub = Subscription(thread=t, stop_event=stop)
        assert sub.active is True
        sub.unsubscribe()
        assert stop.is_set()
        assert sub.active is False


class TestSubscribe:
    def test_subscribe_receives_events(self):
        """subscribe() should deliver every event from stream() to the handler."""
        events = _make_events(3)
        client = EventsClient("https://api.grantex.dev", "test-key")

        with patch.object(client, "stream", return_value=iter(events)):
            received: list[GrantexEvent] = []
            sub = client.subscribe(lambda e: received.append(e))
            # Wait for the background thread to finish consuming the iterator
            sub._thread.join(timeout=2)

        assert received == events

    def test_subscribe_with_options(self):
        """subscribe() should forward StreamOptions to stream()."""
        events = _make_events(1)
        client = EventsClient("https://api.grantex.dev", "test-key")
        opts = StreamOptions(types=["grant.created"])

        with patch.object(client, "stream", return_value=iter(events)) as mock_stream:
            sub = client.subscribe(lambda e: None, options=opts)
            sub._thread.join(timeout=2)
            mock_stream.assert_called_once_with(opts)

    def test_unsubscribe_stops_stream(self):
        """unsubscribe() should stop processing events mid-stream."""
        client = EventsClient("https://api.grantex.dev", "test-key")
        received: list[GrantexEvent] = []
        barrier = threading.Event()

        def slow_generator():
            for evt in _make_events(10):
                yield evt
                # After yielding each event, give the test a chance to unsubscribe
                barrier.wait(timeout=2)
                barrier.clear()

        with patch.object(client, "stream", return_value=slow_generator()):
            sub = client.subscribe(lambda e: received.append(e))
            # Wait for first event to be processed
            for _ in range(50):
                if len(received) >= 1:
                    break
                time.sleep(0.01)
            # Unsubscribe after the first event
            sub.unsubscribe()
            barrier.set()  # unblock generator in case it's waiting

        # Should have received at most a couple of events, not all 10
        assert 1 <= len(received) < 10

    def test_subscribe_error_calls_on_error(self):
        """When stream() raises, the on_error callback should be invoked."""
        client = EventsClient("https://api.grantex.dev", "test-key")
        error_holder: list[Exception] = []

        def failing_stream(options=None):
            raise ConnectionError("SSE connection lost")

        with patch.object(client, "stream", side_effect=failing_stream):
            sub = client.subscribe(
                lambda e: None,
                on_error=lambda exc: error_holder.append(exc),
            )
            sub._thread.join(timeout=2)

        assert len(error_holder) == 1
        assert isinstance(error_holder[0], ConnectionError)
        assert "SSE connection lost" in str(error_holder[0])

    def test_subscribe_error_no_callback_swallowed(self):
        """When stream() raises and no on_error is given, the error is silently swallowed."""
        client = EventsClient("https://api.grantex.dev", "test-key")

        def failing_stream(options=None):
            raise RuntimeError("boom")

        with patch.object(client, "stream", side_effect=failing_stream):
            sub = client.subscribe(lambda e: None)
            sub._thread.join(timeout=2)

        # Thread exited cleanly without propagating
        assert sub.active is False

    def test_subscribe_returns_subscription(self):
        """subscribe() should return a Subscription instance."""
        client = EventsClient("https://api.grantex.dev", "test-key")

        with patch.object(client, "stream", return_value=iter([])):
            sub = client.subscribe(lambda e: None)
            assert isinstance(sub, Subscription)
            sub._thread.join(timeout=2)



"""Tests for PassportsClient."""
from __future__ import annotations

import json

import pytest
import respx
import httpx

from grantex import (
    Grantex,
    IssuePassportParams,
    ListPassportsParams,
    MaxTransactionAmount,
)

MOCK_ISSUED_PASSPORT = {
    "passportId": "urn:grantex:passport:01HXYZ",
    "credential": {
        "@context": [
            "https://www.w3.org/ns/credentials/v2",
            "https://grantex.dev/contexts/mpp/v1",
        ],
        "type": ["VerifiableCredential", "AgentPassportCredential"],
        "id": "urn:grantex:passport:01HXYZ",
    },
    "encodedCredential": "eyJAY29udGV4dCI6WyJodHRwczovL3d3dy53My5vcmcvbnM...",
    "expiresAt": "2026-04-06T00:00:00Z",
}

MOCK_GET_PASSPORT = {
    "status": "active",
    "@context": [
        "https://www.w3.org/ns/credentials/v2",
        "https://grantex.dev/contexts/mpp/v1",
    ],
    "type": ["VerifiableCredential", "AgentPassportCredential"],
    "id": "urn:grantex:passport:01HXYZ",
    "issuer": "did:web:grantex.dev",
}

MOCK_REVOKE_RESPONSE = {
    "revoked": True,
    "revokedAt": "2026-04-05T12:00:00Z",
}


@pytest.fixture
def client() -> Grantex:
    return Grantex(api_key="test-key")


@respx.mock
def test_issue_passport(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/passport/issue").mock(
        return_value=httpx.Response(201, json=MOCK_ISSUED_PASSPORT)
    )
    result = client.passports.issue(
        IssuePassportParams(
            agent_id="ag_01",
            grant_id="grant_01",
            allowed_mpp_categories=["inference", "compute"],
            max_transaction_amount=MaxTransactionAmount(amount=100.0, currency="USDC"),
            payment_rails=["tempo"],
            expires_in="24h",
        )
    )

    assert result.passport_id == "urn:grantex:passport:01HXYZ"
    assert result.encoded_credential.startswith("eyJ")
    assert result.expires_at == "2026-04-06T00:00:00Z"
    assert "AgentPassportCredential" in result.credential.get("type", [])

    body = json.loads(route.calls[0].request.content)
    assert body["agentId"] == "ag_01"
    assert body["grantId"] == "grant_01"
    assert body["allowedMPPCategories"] == ["inference", "compute"]
    assert body["maxTransactionAmount"] == {"amount": 100.0, "currency": "USDC"}
    assert body["paymentRails"] == ["tempo"]
    assert body["expiresIn"] == "24h"


@respx.mock
def test_issue_passport_minimal(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/passport/issue").mock(
        return_value=httpx.Response(201, json=MOCK_ISSUED_PASSPORT)
    )
    result = client.passports.issue(
        IssuePassportParams(
            agent_id="ag_01",
            grant_id="grant_01",
            allowed_mpp_categories=["general"],
            max_transaction_amount=MaxTransactionAmount(amount=50.0, currency="USD"),
        )
    )

    assert result.passport_id == "urn:grantex:passport:01HXYZ"

    body = json.loads(route.calls[0].request.content)
    assert "paymentRails" not in body
    assert "expiresIn" not in body
    assert "parentPassportId" not in body


@respx.mock
def test_get_passport(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/passport/urn%3Agrantex%3Apassport%3A01HXYZ").mock(
        return_value=httpx.Response(200, json=MOCK_GET_PASSPORT)
    )
    result = client.passports.get("urn:grantex:passport:01HXYZ")

    assert result.status == "active"
    assert result.raw["issuer"] == "did:web:grantex.dev"
    assert "AgentPassportCredential" in result.raw["type"]


@respx.mock
def test_revoke_passport(client: Grantex) -> None:
    route = respx.post(
        "https://api.grantex.dev/v1/passport/urn%3Agrantex%3Apassport%3A01HXYZ/revoke"
    ).mock(return_value=httpx.Response(200, json=MOCK_REVOKE_RESPONSE))

    result = client.passports.revoke("urn:grantex:passport:01HXYZ")

    assert result.revoked is True
    assert result.revoked_at == "2026-04-05T12:00:00Z"


@respx.mock
def test_list_passports_with_params(client: Grantex) -> None:
    mock_list = [
        {
            "passportId": "urn:grantex:passport:01HXYZ",
            "credential": {},
            "encodedCredential": "",
            "expiresAt": "2026-04-06T00:00:00Z",
        },
    ]
    respx.get(url__regex=r"/v1/passports\?.*agentId=ag_01").mock(
        return_value=httpx.Response(200, json=mock_list)
    )

    result = client.passports.list(
        ListPassportsParams(agent_id="ag_01", status="active")
    )

    assert len(result.passports) == 1
    assert result.passports[0].passport_id == "urn:grantex:passport:01HXYZ"


@respx.mock
def test_list_passports_without_params(client: Grantex) -> None:
    respx.get("https://api.grantex.dev/v1/passports").mock(
        return_value=httpx.Response(200, json=[])
    )
    result = client.passports.list()
    assert len(result.passports) == 0


@respx.mock
def test_issue_passport_with_parent(client: Grantex) -> None:
    route = respx.post("https://api.grantex.dev/v1/passport/issue").mock(
        return_value=httpx.Response(201, json=MOCK_ISSUED_PASSPORT)
    )
    client.passports.issue(
        IssuePassportParams(
            agent_id="ag_02",
            grant_id="grant_02",
            allowed_mpp_categories=["inference"],
            max_transaction_amount=MaxTransactionAmount(amount=25.0, currency="USDC"),
            parent_passport_id="urn:grantex:passport:PARENT",
        )
    )

    body = json.loads(route.calls[0].request.content)
    assert body["parentPassportId"] == "urn:grantex:passport:PARENT"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Callable, Iterator, Optional, Sequence
import json
import threading

import httpx


EventHandler = Callable[["GrantexEvent"], None]
ErrorHandler = Callable[[Exception], None]


@dataclass(frozen=True)
class GrantexEvent:
    id: str
    type: str
    created_at: str
    data: dict[str, Any]

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> "GrantexEvent":
        return cls(
            id=d["id"],
            type=d["type"],
            created_at=d["createdAt"],
            data=d.get("data", {}),
        )


@dataclass(frozen=True)
class StreamOptions:
    types: Optional[Sequence[str]] = None


class Subscription:
    """Handle returned by ``EventsClient.subscribe`` to control the background stream."""

    def __init__(self, thread: threading.Thread, stop_event: threading.Event) -> None:
        self._thread = thread
        self._stop_event = stop_event

    @property
    def active(self) -> bool:
        """Return ``True`` if the subscription is still running."""
        return self._thread.is_alive()

    def unsubscribe(self) -> None:
        """Stop the background stream and wait for the thread to exit."""
        self._stop_event.set()
        self._thread.join(timeout=5)


class EventsClient:
    def __init__(self, base_url: str, api_key: str) -> None:
        self._base_url = base_url
        self._api_key = api_key

    def stream(self, options: Optional[StreamOptions] = None) -> Iterator[GrantexEvent]:
        """Connect to the SSE event stream. Yields GrantexEvent objects."""
        params = {}
        if options and options.types:
            params["types"] = ",".join(options.types)

        url = f"{self._base_url}/v1/events/stream"
        with httpx.stream(
            "GET",
            url,
            params=params,
            headers={"Authorization": f"Bearer {self._api_key}"},
            timeout=None,
        ) as response:
            response.raise_for_status()
            buffer = ""
            for chunk in response.iter_text():
                buffer += chunk
                while "\n" in buffer:
                    line, buffer = buffer.split("\n", 1)
                    if line.startswith("data: "):
                        try:
                            data = json.loads(line[6:])
                            yield GrantexEvent.from_dict(data)
                        except (json.JSONDecodeError, KeyError):
                            pass

    def subscribe(
        self,
        handler: EventHandler,
        options: Optional[StreamOptions] = None,
        *,
        on_error: Optional[ErrorHandler] = None,
    ) -> Subscription:
        """Subscribe to events with a callback handler.

        Starts a background thread that calls ``stream()`` and invokes
        *handler* for each event.  Returns a :class:`Subscription` whose
        ``unsubscribe()`` method stops the thread.

        Args:
            handler: Called with each :class:`GrantexEvent` received.
            options: Optional :class:`StreamOptions` to filter event types.
            on_error: Optional callback invoked when the stream raises an
                exception.  If not provided, errors are silently swallowed
                and the stream stops.
        """
        stop = threading.Event()

        def _run() -> None:
            try:
                for event in self.stream(options):
                    if stop.is_set():
                        break
                    handler(event)
            except Exception as exc:  # noqa: BLE001
                if on_error is not None:
                    on_error(exc)

        thread = threading.Thread(target=_run, daemon=True)
        thread.start()
        return Subscription(thread=thread, stop_event=stop)
